#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
typedef long LONG;
LONG inv_ff(LONG a);  // Data/rref.c 

void *mymalloc(size_t n);
void myfree(void *p);
void freemat(LONG **p,int row);
void freevect(LONG *p);
void freeintvect(int *p);
LONG **newmat(int row,int col);
LONG *newvect(int n);
int *newintvect(int n);
LONG **matrix_identity_matrix(int n);
void matrix_transpose_nc(LONG **destMat,LONG **mat,int row,int col);
void rowx(LONG **mat,int i,int j,int row,int col);
void rowm(LONG **mat,int i,LONG c,int row,int col);
void rowa(LONG **mat,int i,int j,LONG c,int row,int col);
int is_zero_vect(LONG *V,int n);
void print_mat(LONG **mat,int row, int col);

struct MData {
  LONG **M;
  int *Indep_columns;
  int len;
  LONG **P;
  int row;
  int col;
};

struct MData gauss_elim_ff(LONG **mat,int row,int col,LONG prime);

/* for checking results */
LONG **matrix_mult(LONG **m1,LONG **m2, int ii,int jj,int kk);
LONG **matrix_sub(LONG **m1,LONG **m2,int row,int col);
void print_intvect(int *v,int len);

/* for outputing results */
void fprint_mat_asir_form(FILE *fp2,LONG **m,int row,int col);
void fprint_intvec_asir_form(FILE *fp2,int *v,int n);
  
LONG Prime=13;
int Check=0;
int Debug=1;
size_t Used_memory=0;


void *mymalloc(size_t n) {
  void *m;
  Used_memory += n;
  m = malloc(n);
  if (m == NULL) {fprintf(stderr,"No more memory. Have allocated %zu bytes\n",Used_memory); exit(-1);
  }
  return m;
}
void myfree(void *p) {
  free(p);
}
void freemat(LONG **p,int row) {
  int i;
  for (i=0; i<row; i++) {
      myfree(p[i]);
  }
  myfree(p);
}
void freevect(LONG *p) {
  myfree(p);
}
void freeintvect(int *p) {
  myfree(p);
}


LONG inv_ff(LONG a)
{
  LONG f1,f2,a1,a2,q,r,mod;
  mod = Prime;
  f1 = a; f2 = mod;
  a1 = 1; a2 = 0;
  while ( 1 ) {
    q = f1/f2; r = f1 - f2*q; 
    f1 = f2; f2 = r;
    if ( r == 0 ) 
      break;
    r = (a2*q)%mod;
    r = a1-r;
    if ( r < 0 ) r += mod;
    a1 = a2; a2 = r;
  }
  return a2;
}
LONG **newmat(int row,int col) {
  LONG **m;
  int i;
  m = (LONG **)mymalloc(sizeof(LONG *)*row);
  for (i=0; i<row; i++) m[i]=(LONG *)mymalloc(sizeof(LONG)*col);
  if (Debug) fprintf(stderr,"Have allocated %zu M bytes.\n",Used_memory/1000000);
  return m;
}
LONG *newvect(int n) {
  LONG *v;
  int i;
  v = (LONG *)mymalloc(sizeof(LONG)*n);
  if (Debug) fprintf(stderr,"Have allocated %zu M bytes.\n",Used_memory/1000000);
  return v;
}
int *newintvect(int n) {
  int *v;
  int i;
  v = (int *)mymalloc(sizeof(int)*n);
  if (Debug) fprintf(stderr,"Have allocated %zu M bytes.\n",Used_memory/1000000);
  return v;
}
LONG **matrix_identity_matrix(int n) {
  LONG **m;
  int i,j;
  m = newmat(n,n);
  for (i=0; i<n; i++) for (j=0; j<n; j++) {
      if (i != j) m[i][j]=0; else m[i][j]=1;
    }
  return m;
}
/* _nc: not_clone */
void matrix_transpose_nc(LONG **destMat,LONG **mat,int row,int col) {
  LONG **m;
  int i,j;
  m = destMat;
  for (i=0; i<row; i++) {
    for (j=0; j<col; j++) {
      m[j][i]=mat[i][j];
    }
  }
}
void rowx(LONG **mat,int i,int j,int row,int col) {
  /* exchange i-th row and j-th row. */
  int q;
  LONG t;
  for (q=0; q<col; q++) {
     t=mat[i][q]; mat[i][q]=mat[j][q]; mat[j][q]=t;
  }
}
void rowm(LONG **mat,int i,LONG c,int row,int col) {
  /* multiply i-th row by c. */
  int q;
  LONG t;
  for (q=0; q<col; q++) {
     mat[i][q] = (mat[i][q]*c) % Prime;
  }
}
void rowa(LONG **mat,int i,int j,LONG c,int row,int col) {
  /* i-th row = i-th row + c*(j-th row) */
  int q;
  LONG t;
  for (q=0; q<col; q++) {
     mat[i][q] = (mat[i][q] + ((c*mat[j][q]) % Prime)) % Prime;
  }
}

int is_zero_vect(LONG *v,int n) {
  int i;
  for (i=0; i<n; i++) {
    if (v[i] != 0) return 0;
  }
  return 1;
}

void print_mat(LONG **mat,int row, int col) {
  int i,j;
  for (i=0; i<row; i++) {
    for (j=0; j<col; j++) {
      printf("%11ld ",mat[i][j]);
    }
    printf("\n");
  }
}
LONG **matrix_mult(LONG **m1,LONG **m2, int ii,int jj,int kk) {
  int i,j,k;
  LONG **m;
  LONG t;
  m = newmat(ii,kk);
  for (i=0; i<ii; i++) {
    for (k=0; k<kk; k++) {
      t = 0;
      for (j=0; j<jj; j++) {
	t = (t+(m1[i][j]*m2[j][k] % Prime)) % Prime;
      }
      m[i][k]=t;
    }
  }
  return m;
}
void print_intvect(int *v,int len) {
  int i;
  for (i=0; i<len; i++) printf("%12d ",v[i]);
}
LONG **matrix_sub(LONG **m1,LONG **m2,int row,int col) {
  int i,j;
  LONG **m;
  m = newmat(row,col);
  for (i=0; i<row; i++) {
    for (j=0; j<col; j++) {
      if (m2 != NULL) {
	m[i][j] = (m1[i][j]-m2[i][j]) % Prime;
      }else {
	m[i][j] = m1[i][j] % Prime;
      }
    }
  }
  return m;
}

void fprint_mat_asir_form(FILE *fp2,LONG **m,int row,int col) {
  int i,j;
  fprintf(fp2,"[");
  for (i=0; i<row; i++) {
    fprintf(fp2," [");
    for (j=0; j<col; j++) {
      fprintf(fp2,"%ld",m[i][j]);
      if (j != col-1) fprintf(fp2,",");
      else fprintf(fp2,"]");
    }
    if (i != row-1) fprintf(fp2,",");
    else fprintf(fp2,"]");
  }
}
void fprint_intvec_asir_form(FILE *fp2,int *v,int n) {
  int j;
  fprintf(fp2,"[");
  for (j=0; j<n; j++) {
    fprintf(fp2,"%d",v[j]);
    if (j != n-1) fprintf(fp2,",");
    else fprintf(fp2,"]");
  }
}


struct MData gauss_elim_ff(LONG **M,int row,int col,LONG myprime) {
  int RL,CL;
  LONG **P;
  int Current_row,Current_column,I,J,K,index;
  LONG **MT;
  int *Indep_columns;
  struct MData md;
  Prime=myprime;
  RL=row;
  CL=col;
  P=matrix_identity_matrix(RL);
  Current_row=0;
  Current_column=0;
  Indep_columns=newintvect(col); index=0;
  MT=newmat(col,row);
  while((Current_row<RL)&&(Current_column<CL))
    {
      matrix_transpose_nc(MT,M,row,col);
      J=Current_column;
      while((J<CL)&&(is_zero_vect(MT[J],row))){
	J++;
      }
      if(J<CL){
	I=Current_row;
	while((I<RL)&&(M[I][J]==0)){
	  I++;
	}
	if(I<RL){
	  rowx(P,Current_row,I,row,row);
	  rowx(M,Current_row,I,row,col);
	  rowm(P,Current_row,inv_ff(M[Current_row][J]),row,row);
	  rowm(M,Current_row,inv_ff(M[Current_row][J]),row,col);
	  for(K=0;K<RL;K++){
	    if((K!=Current_row)&&(M[K][J]!=0)){
	      rowa(P,K,Current_row,-M[K][J],row,row);
	      rowa(M,K,Current_row,-M[K][J],row,col);
	    }
	  }
	  Current_row++;
	  Indep_columns[index++]=J;
	}
      }
      Current_column=J+1;
    }
 md.M=M;
 md.Indep_columns=Indep_columns;
 md.len=index;
 md.P=P;
 md.row=row;
 md.col=col;

 freemat(MT,row);

 return md;
}


int main_test1() {
  LONG **m;
  LONG **m_orig;
  int row,col;
  struct MData md;
  row=3; col=2;
  m = newmat(row,col);
  m[0][0]=1; m[0][1]=2;
  m[1][0]=3; m[1][1]=4;
  m[2][0]=4; m[2][1]=6;
  m_orig=matrix_sub(m,NULL,row,col);
  print_mat(m,row,col); printf("\n");
  md=gauss_elim_ff(m,row,col,Prime);
  printf("m=\n");
  print_mat(m,row,col);
  printf("independent col=\n");
  print_intvect(md.Indep_columns,md.len); printf("\n");
  printf("P=\n");
  print_mat(md.P,row,row);

  printf("Should be 0=\n");
  print_mat(matrix_sub(m,matrix_mult(md.P,m_orig,row,row,col),row,col),row,col);
  return 0;
}

void usage() {
  printf("gauss_elim_ff file_of_matrix [file_to_output]\n");
}
/*

*/
int main(int argc, char *argv[]) {
  FILE *fp;
  FILE *fp2;
  int row,col,i,j;
  LONG **m;
  LONG **m_orig;
  LONG **diff;
  struct MData md;
  fp = fopen(argv[1],"r");
  if (fp==NULL) {
    fprintf(stderr,"Error: file %s is not found.\n",argv[1]);
    return -1;
  }
  if (argc==2) fp2=stdout;
  else {
    if (argc < 3) {
      usage(); return -1;
    }
    fp2 = fopen(argv[2],"w");
  }
  fscanf(fp,"%ld",&Prime);
  fscanf(fp,"%d",&row); fscanf(fp,"%d",&col);
  m = newmat(row,col);
  for (i=0; i<row; i++) for (j=0; j<col; j++) fscanf(fp,"%ld",&(m[i][j]));

  m_orig=matrix_sub(m,NULL,row,col);
  md=gauss_elim_ff(m,row,col,Prime);

  fprintf(fp2,"NT_MData=[");
  fprint_mat_asir_form(fp2,m,row,col);
  fprintf(fp2,", ");
  fprint_intvec_asir_form(fp2,md.Indep_columns,md.len);
  fprintf(fp2,", ");
  fprint_mat_asir_form(fp2,md.P,row,row);
  fprintf(fp2,"]; ");
  fclose(fp2);

  if (Check) {
    /* diff should be 0 */
    LONG **tm;
    diff=matrix_sub(m,tm=matrix_mult(md.P,m_orig,row,row,col),row,col);
    for (i=0; i<row; i++) for (j=0; j<col; j++) {
	if (diff[i][j] != 0) {
	  fprintf(stderr,"Error of diff check\n"); return(-1);
	}
      }
    freemat(diff,row); freemat(tm,row);
  }
  freemat(m,row);
  freemat(m_orig,row);
  freemat(md.P,row);
  freeintvect(md.Indep_columns);
  return 0;
}
